const { default: axios } = require("axios");
const { ethers, BigNumber } = require("ethers");

import { MultiCall } from '@indexed-finance/multicall';

import { TransferType } from "../constants/index";
import * as contracts from "../abis/index";
import { config, network, uriToHttp } from "../config/nets";

class RouterSDK {
    constructor(signerOrProvider, optionalParams) {
        return (async () => {

            // All async code here
            this.signerOrProvider = signerOrProvider;

            this.chainId = this.signerOrProvider.provider.network.chainId;
            this.config = config[this.chainId];

            if (this.config.contracts.multicall && this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
                this.multicallContract = new ethers.Contract(this.config.contracts.multicall, contracts.MulticallAbi, this.signerOrProvider)
            }

            this.setParams(optionalParams || {});

            this.erc20RouterContract = new ethers.Contract(this.ERC20RouterAddress, this.ERC20RouterAbi, this.signerOrProvider);
            this.legacyRouterContract = new ethers.Contract(this.LegacyRouterAddress, this.LegacyRouterAbi, this.signerOrProvider);
            this.legacyRouterProxyContract = new ethers.Contract(this.LegacyRouterProxyAddress, this.LegacyRouterProxyAbi, this.signerOrProvider);

            let validatorArray = [];
            let initialValidatorVersion;
            let initialNFTValidatorVersion;

            try {
                initialValidatorVersion = (await axios.get(`${this.validatorURL}/status`)).data.version;
                if (this.frontEndVersion !== 1) {
                    initialNFTValidatorVersion = (await axios.get(`${this.nftValidatorURL}/status`)).data.version;
                }
            } catch (e) {
                initialValidatorVersion = await this._getDefaultConfiguredValidatorVersion();
                if (this.frontEndVersion !== 1) {
                    initialNFTValidatorVersion = await this._getDefaultConfiguredNftValidatorVersion();
                }
                if (!initialValidatorVersion || !initialNFTValidatorVersion) {
                    throw new Error('Validators are not accessible')
                }
            }

            const initialValidatorMajorVersion = parseInt(initialValidatorVersion.split('.')[0]);
            const initialValidatorMinorVersion = parseInt(initialValidatorVersion.split('.')[1]);
            let initialNftValidatorMajorVersion;
            let initialNftValidatorMinorVersion;
            if (this.frontEndVersion !== 1) {
                initialNftValidatorMajorVersion = parseInt(initialNFTValidatorVersion.split('.')[0]);
                initialNftValidatorMinorVersion = parseInt(initialNFTValidatorVersion.split('.')[1]);
            }
            const optionalParamsValidatorMajorVersion = parseInt(optionalParams["validatorTargetVersion"].split('.')[0]);
            const optionalParamsValidatorMinorVersion = parseInt(optionalParams["validatorTargetVersion"].split('.')[1]);
            let optionalParamsNftValidatorMajorVersion;
            let optionalParamsNftValidatorMinorVersion;
            if (this.frontEndVersion !== 1) {
                optionalParamsNftValidatorMajorVersion = parseInt(optionalParams["nftValidatorTargetVersion"].split('.')[0]);
                optionalParamsNftValidatorMinorVersion = parseInt(optionalParams["nftValidatorTargetVersion"].split('.')[1]);
            }

            this.axios = new axios.create({ baseURL: this.validatorURL });
            this.nftAxios = new axios.create({ baseURL: this.nftValidatorURL });
            this.backendAxios = new axios.create({ baseURL: this.abridgeBackendURL });
            this.validatorVersion = initialValidatorVersion;
            this.nftValidatorVersion = initialNFTValidatorVersion;

            // Check for passed validatorTargetVersion
            if ((optionalParams && optionalParams["validatorTargetVersion"]) && (initialValidatorMajorVersion !== optionalParamsValidatorMajorVersion || !(initialValidatorMinorVersion >= optionalParamsValidatorMinorVersion))) {

                if (optionalParams && optionalParams.validators) {
                    validatorArray = this._getAllValidators(optionalParams.validators)
                } else {
                    validatorArray = this._getAllValidators(this.config.validators)
                }
                await this._setValidatorBasedOnVersion(validatorArray, optionalParams["validatorTargetVersion"])
            }

            // Check for passed nftValidatorTargetVersion
            if ((this.frontEndVersion !== 1 && optionalParams && optionalParams["nftValidatorTargetVersion"]) && (initialNftValidatorMajorVersion !== optionalParamsNftValidatorMajorVersion || !(initialNftValidatorMinorVersion >= optionalParamsNftValidatorMinorVersion))) {

                if (optionalParams && optionalParams.nftValidators) {
                    validatorArray = this._getAllValidators(optionalParams.nftValidators)
                } else {
                    validatorArray = this._getAllValidators(this.config.nftValidators)
                }
                await this._setValidatorBasedOnVersion(validatorArray, optionalParams["nftValidatorTargetVersion"])
            }

            return this; // when done
        })();
    }

    /**
     * @param optionalParams
     * @property legacyRouterABI: legacyErc20/IRouterDiamond 
     * @property legacyRouterAddress: address 
     * @property legacyRouterProxyABI: legacyErc20/RouterProxy
     * @property legacyRouterProxyAddress: address
     * @property wrappedErc721ABI: WrappedERC721
     * @property teleportABI: ITeleport
     * @property nftRouterABI: NFTIRouterDiamond
     * @property erc20RouterABI: ERC20IRouterDiamond
     * @property erc20RouterAddress: address 
     * @property erc20PermitABI: ERC20Permit
     * @property validators: object
     * @property threshold: number 
     * @property validatorTargetVersion: string
     * @property nftValidators: string
     * @property nftValidatorTargetVersion: string
     * @property abridgeBackend: string
     * @property frontEndVersion: number

     */
    setParams(optionalParams) {
        if (optionalParams["legacyRouterABI"]) {
            this.LegacyRouterAbi = optionalParams.legacyRouterABI;
        } else {
            this.LegacyRouterAbi = contracts.LegacyRouterAbi;
        }

        if (optionalParams["legacyRouterAddress"]) {
            this.LegacyRouterAddress = optionalParams.legacyRouterAddress;
        } else {
            this.LegacyRouterAddress = this.config.contracts.legacyErc20['router'];
        }

        if (optionalParams["legacyRouterProxyABI"]) {
            this.LegacyRouterProxyAbi = optionalParams.legacyRouterProxyABI;
        } else {
            this.LegacyRouterProxyAbi = contracts.LegacyRouterProxyAbi;
        }

        if (optionalParams["legacyRouterProxyAddress"]) {
            this.LegacyRouterProxyAddress = optionalParams.legacyRouterProxyAddress;
        } else {
            this.LegacyRouterProxyAddress = this.config.contracts.legacyErc20['router-proxy'];
        }

        if (optionalParams["wrappedErc721ABI"]) {
            this.WrappedERC721Abi = optionalParams.wrappedErc721ABI;
        } else {
            this.WrappedERC721Abi = contracts.WrappedERC721Abi;
        }

        if (optionalParams["teleportABI"]) {
            this.ITeleportAbi = optionalParams.teleportABI;
        } else {
            this.ITeleportAbi = contracts.ITeleportAbi;
        }

        if (optionalParams["nftRouterABI"]) {
            this.NFTRouterAbi = optionalParams.nftRouterABI;
        } else {
            this.NFTRouterAbi = contracts.NFTRouterAbi;
        }

        if (optionalParams["erc20RouterABI"]) {
            this.ERC20RouterAbi = optionalParams.erc20RouterABI;
        } else {
            this.ERC20RouterAbi = contracts.ERC20RouterAbi;
        }

        if (optionalParams["erc20RouterAddress"]) {
            this.ERC20RouterAddress = optionalParams.erc20RouterAddress;
        } else {
            this.ERC20RouterAddress = this.config.contracts.erc20Router;
        }

        if (optionalParams["erc20PermitABI"]) {
            this.ERC20PermitAbi = optionalParams.erc20PermitABI;
        } else {
            this.ERC20PermitAbi = contracts.ERC20PermitAbi;
        }

        if (optionalParams["validators"]) {
            this.validatorURL = this.getRandomProperty(optionalParams.validators);
        } else {
            this.validatorURL = this.getRandomProperty(this.config.validators);
        }

        let threshold;
        if (optionalParams["threshold"]) {
            threshold = ethers.utils.formatUnits(optionalParams.threshold)
        } else {
            threshold = ethers.utils.formatUnits(this.config.threshold)
        }

        if (optionalParams["nftValidators"]) {
            this.nftValidatorURL = this.getRandomProperty(optionalParams.nftValidators);
        } else {
            this.nftValidatorURL = this.getRandomProperty(this.config.nftValidators);
        }

        if (optionalParams["abridgeBackend"]) {
            this.abridgeBackendURL = optionalParams.abridgeBackend;
        } else {
            this.abridgeBackendURL = this.config.abridgeBackend;
        }

        if (optionalParams["frontEndVersion"]) {
            this.frontEndVersion = optionalParams.frontEndVersion;
        } else {
            this.frontEndVersion = 2;
        }


        this.tokenAllowanceThreshold = ethers.utils.parseUnits(threshold);
    }

    setValidator(url) {
        this.validatorURL = url;
        this.axios = new axios.create({ baseURL: this.validatorURL });
    }

    setNftValidator(url) {
        this.nftValidatorURL = url;
        this.nftAxios = new axios.create({ baseURL: this.nftValidatorURL });
    }

    getRandomProperty(obj) {
        const keys = Object.keys(obj);
        const randomIndex = keys.length * Math.random() << 0;

        return obj[keys[randomIndex]];
    }

    async sendLegacy(targetChainId, tokenAddress, amount, receiver, feeTokenAddress, bridgedTokenIsNativeToken) {
        const nativeToken = await this.legacyRouterContract.wrappedToNativeToken(tokenAddress);
        if (feeTokenAddress) {
            // Check of fee token is native token
            let fee = feeTokenAddress === this.legacyRouterProxyContract.address ? await this.getProxyFeeLegacy(feeTokenAddress) : 0;

            // Check if bridged token is native token
            if (bridgedTokenIsNativeToken) {
                fee = amount.add(fee);
            }
            if (nativeToken.chainId == 0) {
                return await this.legacyRouterProxyContract.populateTransaction.lock(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, { value: fee });
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.legacyRouterProxyContract.populateTransaction.burnAndTransfer(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, { value: fee })
            }

            return await this.legacyRouterProxyContract.populateTransaction.burn(feeTokenAddress, tokenAddress, amount, receiver, { value: fee })

        } else {
            if (nativeToken.chainId == 0) {
                return await this.legacyRouterContract.populateTransaction.lock(targetChainId, tokenAddress, amount, receiver);
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.legacyRouterContract.populateTransaction.burnAndTransfer(targetChainId, tokenAddress, amount, receiver);
            }

            return await this.legacyRouterContract.populateTransaction.burn(tokenAddress, amount, receiver);
        }
    }

    async sendWithPermitLegacy(targetChainId, tokenAddress, amount, receiver, v, r, s, deadline, feeTokenAddress, bridgedTokenIsNativeToken) {
        const nativeToken = await this.legacyRouterContract.wrappedToNativeToken(tokenAddress);
        if (feeTokenAddress) {
            // Check of fee token is native token
            let fee = feeTokenAddress === this.legacyRouterProxyContract.address ? await this.getProxyFeeLegacy(feeTokenAddress) : 0;

            // Check if bridged token is native token
            if (bridgedTokenIsNativeToken) {
                fee = amount.add(fee);
            }
            if (nativeToken.chainId == 0) {
                return await this.legacyRouterProxyContract.populateTransaction.lockWithPermit(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, deadline, v, r, s, { value: fee });
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.legacyRouterProxyContract.populateTransaction.burnAndTransferWithPermit(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, deadline, v, r, s, { value: fee });
            }

            return await this.legacyRouterProxyContract.populateTransaction.burnWithPermit(feeTokenAddress, tokenAddress, amount, receiver, deadline, v, r, s, { value: fee });
        } else {
            if (nativeToken.chainId == 0) {
                return await this.legacyRouterContract.populateTransaction.lockWithPermit(targetChainId, tokenAddress, amount, receiver, deadline, v, r, s);
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.legacyRouterContract.populateTransaction.burnAndTransferWithPermit(targetChainId, tokenAddress, amount, receiver, deadline, v, r, s);
            }

            return await this.legacyRouterContract.populateTransaction.burnWithPermit(tokenAddress, amount, receiver, deadline, v, r, s);
        }

    }

    async claimLegacy(transfer) {
        switch (transfer.transferType) {
            case TransferType.LOCK:
                return await this.legacyRouterContract.populateTransaction.mint(
                    transfer.nativeChainId === 0 ? transfer.sourceChainId : transfer.nativeChainId,
                    transfer.nativeToken,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures),
                    {
                        name: transfer.wrappedTokenName,
                        symbol: transfer.wrappedTokenSymbol,
                        decimals: transfer.wrappedTokenDecimals
                    }
                );
            case TransferType.BURN:
                return await this.legacyRouterContract.populateTransaction.unlock(
                    transfer.sourceChainId,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.nativeToken,
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures)
                );
            case TransferType.BURN_AND_TRANSFER:
                return await this.legacyRouterContract.populateTransaction.mint(
                    transfer.nativeChainId,
                    transfer.nativeToken,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures),
                    {
                        name: transfer.wrappedTokenName,
                        symbol: transfer.wrappedTokenSymbol,
                        decimals: transfer.wrappedTokenDecimals
                    }
                );
        }

        return null;
    }

    async sendErc20(
        /* control */
        estimateOnly,           // boolean - true if we need to call estimate gas only
        /* egress mandatory-ish */
        targetChainId,          // number
        tokenAddress,           // address - could be AddressZero when bridging native chain currency
        amount,                 // ethers.BigNumber
        receiverAddress,        // address
        feeTokenAddress,        // address - could be AddressZero
        /* optional - auto-delivery */
        deliveryFeeEstimate,    // ethers.BigNumber
        deliveryFeeExpiry,      // number
        deliveryFeeSignature,   // hex string
        /* optional - permit */
        v, r, s, deadline,      // all should be 'undefined' if not using egressWithPermit
    ) {
        let msgValue = 
            !tokenAddress || tokenAddress === ethers.constants.AddressZero 
                ? amount 
                : ethers.BigNumber.from(0);

        if (!feeTokenAddress || feeTokenAddress === ethers.constants.AddressZero) {
            msgValue.add(await this.getServiceFee());
        }

        let deliveryFeeData;
        if (deliveryFeeEstimate && !deliveryFeeEstimate.isZero()) {
            deliveryFeeData = [
                deliveryFeeEstimate,
                deliveryFeeExpiry,
                deliveryFeeSignature
            ];

            msgValue.add(deliveryFeeEstimate);
        } else {
            deliveryFeeData = [0, 0, []];
        }

        usePermit = v || r || s || deadline;

        if (estimateOnly) {
            if (usePermit) {
                return this.erc20RouterContract.estimateGas.egressWithPermit(
                    targetChainId,
                    feeTokenAddress,
                    tokenAddress,
                    amount,
                    receiverAddress,
                    deliveryFeeData,
                    deadline, v, r, s,
                    { value: msgValue });
            } else {
                return this.erc20RouterContract.estimateGas.egress(
                    targetChainId,
                    feeTokenAddress,
                    tokenAddress,
                    amount,
                    receiverAddress,
                    deliveryFeeData,
                    { value: msgValue });
            }
        } else {
            if (usePermit) {
                return this.erc20RouterContract.populateTransaction.egressWithPermit(
                    targetChainId,
                    feeTokenAddress,
                    tokenAddress,
                    amount,
                    receiverAddress,
                    deliveryFeeData,
                    deadline, v, r, s,
                    { value: msgValue });
            } else {
                return this.erc20RouterContract.populateTransaction.egress(
                    targetChainId,
                    feeTokenAddress,
                    tokenAddress,
                    amount,
                    receiverAddress,
                    deliveryFeeData,
                    { value: msgValue });
            }
        }
    }

    async balanceOfMulticall(tokenAddresses, userAddress) {
        const tokenInterface = new ethers.utils.Interface(this.ERC20PermitAbi);
        const fragment_balance = tokenInterface.getFunction('balanceOf');
        const multi = new MultiCall(this.signerOrProvider);
        const inputs = [];
        const outputs = [];

        if (this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
            let result;
            for (let tokenAddress in tokenAddresses) {
                inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_balance, [userAddress]) })
            }
            try {
                result = await this.multicallContract.callStatic.tryBlockAndAggregate(false, inputs);
            } catch (e) {
                console.log(e)
            }
            for (let index = 0; index < result.returnData.length; index++) {
                const balance = tokenInterface.decodeFunctionResult(fragment_balance, result.returnData[index].returnData)
                outputs.push(balance);
            }
            return outputs
        } else {
            for (let tokenAddress in tokenAddresses) {
                inputs.push({ target: tokenAddresses[tokenAddress], function: 'balanceOf', args: [userAddress] })
            }

            const tokenData = await multi.multiCall(this.ERC20PermitAbi, inputs);

            for (let i = 0; i < inputs.length; i++) {
                outputs[i] = tokenData[1][i]
            }
            return outputs
        }
    }

    async balanceOfMulticallV2Legacy(tokenAddresses, userAddress) {
        const tokenInterface = new ethers.utils.Interface(this.ERC20PermitAbi);
        const fragment_balance = tokenInterface.getFunction('balanceOf');
        const multi = new MultiCall(this.signerOrProvider);
        const inputs = [];
        const outputs = [];
        if (this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
            for (let tokenAddress in tokenAddresses) {
                if (tokenAddresses[tokenAddress] === this.legacyRouterProxyContract.address) {
                    const balance = await this.signerOrProvider.getBalance();
                    outputs[tokenAddress] = balance;
                } else {
                    inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_balance, [userAddress]) })
                }
            }

            const tokenData = await this.multicallContract.callStatic.tryBlockAndAggregate(false, inputs);

            for (let i = 0; i < inputs.length; i++) {
                outputs[i] = tokenInterface.decodeFunctionResult(fragment_balance, tokenData.returnData[i].returnData)[0]
            }
            return outputs
        } else {
            for (let tokenAddress in tokenAddresses) {
                if (tokenAddresses[tokenAddress] === this.legacyRouterProxyContract.address) {
                    const balance = await this.signerOrProvider.getBalance();
                    outputs[tokenAddress] = balance;
                } else {
                    inputs.push({ target: tokenAddresses[tokenAddress], function: 'balanceOf', args: [userAddress] })
                }
            }

            const tokenData = await multi.multiCall(this.ERC20PermitAbi, inputs);

            for (let i = 0; i < inputs.length; i++) {
                outputs[i] = tokenData[1][i]
            }
            return outputs
        }
    }

    async balanceOfLegacy(tokenAddress, userAddress) {
        let balance;
        let tokenDecimals;
        if (tokenAddress === this.legacyRouterProxyContract.address) {
            balance = await this.signerOrProvider.getBalance();
            tokenDecimals = await this.legacyRouterProxyContract.decimals();
        } else {
            const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
            balance = await tokenContract.balanceOf(userAddress);
            tokenDecimals = await tokenContract.decimals();
        }

        return { balance, tokenDecimals };
    }

    async balanceOf(tokenAddress, userAddress) {
        let balance;
        let tokenDecimals;

        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        balance = await tokenContract.balanceOf(userAddress);
        tokenDecimals = await tokenContract.decimals();

        return { balance, tokenDecimals };
    }

    async allowanceLegacy(tokenAddress, userAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        const tokenDecimals = await tokenContract.decimals();
        const allowance = await tokenContract.allowance(userAddress, this.legacyRouterContract.address);
        return { allowance, tokenDecimals };
    }

    async allowance(tokenAddress, userAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        const tokenDecimals = await tokenContract.decimals();
        const allowance = await tokenContract.allowance(userAddress, this.erc20RouterContract.address);
        return { allowance, tokenDecimals };
    }

    async isTokenApprovedLegacy(tokenAddress, userAddress, bridgedTokenIsNativeTokenOrNonDefaultFee, threshold /** optional */) {
        const allowanceThreshold = threshold || this.tokenAllowanceThreshold;
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        const currentAllowance = await tokenContract.allowance(userAddress, bridgedTokenIsNativeTokenOrNonDefaultFee ? this.legacyRouterProxyContract.address : this.legacyRouterContract.address);
        return currentAllowance.gte(allowanceThreshold);
    }

    async isTokenApproved(tokenAddress, userAddress, threshold /** optional */) {
        const allowanceThreshold = threshold || this.tokenAllowanceThreshold;
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        const currentAllowance = await tokenContract.allowance(userAddress, this.erc20RouterContract.address);
        return currentAllowance.gte(allowanceThreshold);
    }

    async approveLegacy(tokenAddress, amount, bridgedTokenIsNativeTokenOrNonDefaultFee) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return await tokenContract.populateTransaction.approve(bridgedTokenIsNativeTokenOrNonDefaultFee ? this.legacyRouterProxyContract.address : this.legacyRouterContract.address, amount);
    }

    async approve(tokenAddress, amount) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return await tokenContract.populateTransaction.approve(this.erc20RouterContract.address, amount);
    }

    async approveMaxIntLegacy(tokenAddress, bridgedTokenIsNativeTokenOrNonDefaultFee) {
        return this.approveLegacy(tokenAddress, ethers.constants.MaxUint256, bridgedTokenIsNativeTokenOrNonDefaultFee);
    }

    async approveMaxInt(tokenAddress) {
        return this.approve(tokenAddress, ethers.constants.MaxUint256);
    }

    async getServiceFeeLegacy() {
        return await this.legacyRouterContract.serviceFee();
    }

    async getServiceFee() {
        return await this.erc20RouterContract.serviceFee();
    }

    async getExternalFeeLegacy() {
        return await this.legacyRouterContract.externalFee();
    }

    async getClaimsLegacy(userAddress, targetChainId, page = 'undefined') {
        const transfers = await this.axios(`/getTransfers?targetChainId=${targetChainId}&recipient=${userAddress}&page=${page}`, { responseType: "json" });
        return { claims: transfers.data.transfers, totalCount: transfers.headers['x-total-count'] }
    }

    async getValidatorVersionLegacy() {
        const versionRequest = await this.axios('/status', { responseType: "json" });
        return versionRequest.data.version;
    }

    async getWrappedTokensLegacy(chainId) {
        const wrappedTokens = await this.axios(`/getWrappedTokens?chainId=${chainId}`, { responseType: "json" });
        return wrappedTokens.data.wrappedTokens;
    }

    async isClaimedMulticallLegacy(transfers) {
        const size = 20;
        const arrayOfTransferArrays = [];
        for (let i = 0; i < transfers.length; i += size) {
            arrayOfTransferArrays.push(transfers.slice(i, i + size));
        }

        const multi = new MultiCall(this.signerOrProvider);
        const routerInterface = new ethers.utils.Interface(this.LegacyRouterAbi);
        const fragment = routerInterface.getFunction('hashesUsed');
        let results = [];

        for (let index = 0; index < arrayOfTransferArrays.length; index++) {
            const transferChunkArray = arrayOfTransferArrays[index];
            let inputs = [];

            for (let index = 0; index < transferChunkArray.length; index++) {
                const transfer = transferChunkArray[index];
                let transferChainId = transfer.transferType == TransferType.BURN ? transfer.sourceChainId : transfer.nativeChainId
                let encodedData;
                switch (transfer.transferType) {
                    case TransferType.BURN:
                        encodedData = ethers.utils.defaultAbiCoder.encode(
                            [
                                "uint8",
                                "uint8",
                                "bytes",
                                "address",
                                "uint256",
                                "bytes"
                            ],
                            [
                                transfer.sourceChainId,
                                transfer.targetChainId,
                                ethers.utils.toUtf8Bytes(transfer.sourceTx),
                                transfer.recipient,
                                transfer.amount,
                                transfer.nativeToken
                            ]
                        );
                        break;
                    case TransferType.LOCK:
                        // Backwards compatibility fix
                        if (transfer.nativeChainId === 0) {
                            transfer.nativeChainId = transfer.sourceChainId;
                        }
                        encodedData = ethers.utils.defaultAbiCoder.encode(
                            [
                                "uint8",
                                "uint8",
                                "bytes",
                                "address",
                                "uint256",
                                "bytes",
                                "string",
                                "string",
                                "uint8"
                            ],
                            [
                                transfer.nativeChainId,
                                transfer.targetChainId,
                                ethers.utils.toUtf8Bytes(transfer.sourceTx),
                                transfer.recipient,
                                transfer.amount,
                                transfer.nativeToken,
                                transfer.wrappedTokenName,
                                transfer.wrappedTokenSymbol,
                                transfer.wrappedTokenDecimals
                            ]
                        );
                        break;
                    case TransferType.BURN_AND_TRANSFER:
                        encodedData = ethers.utils.defaultAbiCoder.encode(
                            [
                                "uint8",
                                "uint8",
                                "bytes",
                                "address",
                                "uint256",
                                "bytes",
                                "string",
                                "string",
                                "uint8"
                            ],
                            [
                                transfer.nativeChainId,
                                transfer.targetChainId,
                                ethers.utils.toUtf8Bytes(transfer.sourceTx),
                                transfer.recipient,
                                transfer.amount,
                                transfer.nativeToken,
                                transfer.wrappedTokenName,
                                transfer.wrappedTokenSymbol,
                                transfer.wrappedTokenDecimals
                            ]
                        );
                        break;
                }

                const hashMsg = ethers.utils.keccak256(encodedData);
                const hashData = ethers.utils.arrayify(hashMsg);
                const hashed = ethers.utils.hashMessage(hashData);
                inputs.push([transferChainId, hashed])
            }

            const callArray = [];

            // TODO: Add mainnet Moonbeam chainId
            if (this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
                const statuses = [];
                for (let index = 0; index < transferChunkArray.length; index++) {
                    const callData = routerInterface.encodeFunctionData(fragment, inputs[index])
                    callArray.push({
                        target: this.LegacyRouterAddress,
                        callData: callData
                    })
                }

                const result = await this.multicallContract.callStatic.tryBlockAndAggregate(false, callArray);
                for (let index = 0; index < result.returnData.length; index++) {
                    const status = routerInterface.decodeFunctionResult(fragment, result.returnData[index].returnData)
                    statuses.push(status[0]);
                }
                results.push(...statuses);
            } else {
                for (let index = 0; index < transferChunkArray.length; index++) {
                    callArray.push({
                        target: this.LegacyRouterAddress,
                        function: 'hashesUsed',
                        args: inputs[index]
                    })
                }

                const statuses = await multi.multiCall(this.LegacyRouterAbi, callArray);
                results.push(...statuses[1])
            }
        }
        return results;

    }

    async isClaimedLegacy(transfer) {
        let encodedData;
        switch (transfer.transferType) {
            case TransferType.BURN:
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes"
                    ],
                    [
                        transfer.sourceChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken
                    ]
                );
                break;
            case TransferType.LOCK:
                // Backwards compatibility fix
                if (transfer.nativeChainId === 0) {
                    transfer.nativeChainId = transfer.sourceChainId;
                }
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes",
                        "string",
                        "string",
                        "uint8"
                    ],
                    [
                        transfer.nativeChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken,
                        transfer.wrappedTokenName,
                        transfer.wrappedTokenSymbol,
                        transfer.wrappedTokenDecimals
                    ]
                );
                break;
            case TransferType.BURN_AND_TRANSFER:
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes",
                        "string",
                        "string",
                        "uint8"
                    ],
                    [
                        transfer.nativeChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken,
                        transfer.wrappedTokenName,
                        transfer.wrappedTokenSymbol,
                        transfer.wrappedTokenDecimals
                    ]
                );
                break;
        }

        const hashMsg = ethers.utils.keccak256(encodedData);
        const hashData = ethers.utils.arrayify(hashMsg);

        return await this.legacyRouterContract.hashesUsed(
            transfer.transferType == TransferType.BURN ? transfer.sourceChainId : transfer.nativeChainId,
            ethers.utils.hashMessage(hashData)
        );
    }

    async getTokenName(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return await tokenContract.name();
    }

    async getTokenSymbol(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return await tokenContract.symbol();
    }

    async getTokenDecimals(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return await tokenContract.decimals();
    }

    async getTokenDetailsMulticall(tokenAddresses, userAddress) {
        const tokenInterface = new ethers.utils.Interface(this.ERC20PermitAbi);
        const fragment_balance = tokenInterface.getFunction('balanceOf');
        const fragment_name = tokenInterface.getFunction('name');
        const fragment_symbol = tokenInterface.getFunction('symbol');
        const fragment_decimals = tokenInterface.getFunction('decimals');
        const multi = new MultiCall(this.signerOrProvider);
        const inputs = [];
        const inputs_second_chunk = [];
        const outputs = [];
        const tokenAddressesArray = [];
        try {
            if (this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
                for (let tokenAddress in tokenAddresses) {
                    inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_balance, [userAddress]) })
                    inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_name, []) });
                    inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_symbol, []) })
                    inputs.push({ target: tokenAddresses[tokenAddress], callData: tokenInterface.encodeFunctionData(fragment_decimals, []) })
                    outputs.push({ address: tokenAddresses[tokenAddress] })
                }

                const result = await this.multicallContract.callStatic.tryBlockAndAggregate(false, inputs);
                for (let index = 0; index < result.returnData.length; index = index + 4) {
                    const balance = tokenInterface.decodeFunctionResult(fragment_balance, result.returnData[index].returnData)
                    const name = tokenInterface.decodeFunctionResult(fragment_name, result.returnData[index + 1].returnData)
                    const symbol = tokenInterface.decodeFunctionResult(fragment_symbol, result.returnData[index + 2].returnData)
                    const decimals = tokenInterface.decodeFunctionResult(fragment_decimals, result.returnData[index + 3].returnData)
                    outputs[index / 4].balance = balance[0];
                    outputs[index / 4].name = name[0];
                    outputs[index / 4].symbol = symbol[0];
                    outputs[index / 4].decimals = decimals[0];
                }
                return outputs
            } else {
                for (let tokenAddress in tokenAddresses) {
                    inputs.push({ target: tokenAddresses[tokenAddress], function: 'symbol', args: [] })
                    inputs.push({ target: tokenAddresses[tokenAddress], function: 'decimals', args: [] })
                    inputs_second_chunk.push({ target: tokenAddresses[tokenAddress], function: 'name', args: [] });
                    outputs.push({ address: tokenAddresses[tokenAddress] })
                    tokenAddressesArray.push(tokenAddresses[tokenAddress])
                }
                const balances = await multi.getBalances(tokenAddressesArray, userAddress);
                const tokenData = await multi.multiCall(this.ERC20PermitAbi, inputs);
                const names = await multi.multiCall(this.ERC20PermitAbi, inputs_second_chunk)
                for (let i = 0; i < outputs.length; i++) {
                    outputs[i].balance = balances[1][outputs[i].address];
                    outputs[i].name = names[1][i]
                    outputs[i].symbol = tokenData[1][i * 2]
                    outputs[i].decimals = tokenData[1][i * 2 + 1]
                }
                return outputs
            }
        } catch (e) {
            console.log(e)
        }

    }

    async getTokenDetails(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.ERC20PermitAbi, this.signerOrProvider);
        return {
            name: await tokenContract.name(),
            symbol: await tokenContract.symbol(),
            decimals: await tokenContract.decimals()
        }
    }

    async getWrappedTokenDetailsLegacy(nativeTokenAddress, nativeChainId) {
        const wrappedTokenAddress = await this.legacyRouterContract.nativeToWrappedToken(nativeChainId, nativeTokenAddress);
        const wrappedTokenDetails = await this.getTokenDetails(wrappedTokenAddress);
        return { address: wrappedTokenAddress, ...wrappedTokenDetails };
    }

    async getWrappedTokenDetails(nativeTokenAddress, nativeChainId) {
        const wrappedTokenAddress = await this.erc20RouterContract.nativeToWrappedToken(nativeChainId, nativeTokenAddress);
        const wrappedTokenDetails = await this.getTokenDetails(wrappedTokenAddress);
        return { address: wrappedTokenAddress, ...wrappedTokenDetails };
    }

    async getNativeTokenByWrappedAddressLegacy(wrappedTokenAddress) {
        return await this.legacyRouterContract.wrappedToNativeToken(wrappedTokenAddress);
    }

    async getNativeTokenByWrappedAddress(wrappedTokenAddress) {
        return await this.erc20RouterContract.wrappedToNativeToken(wrappedTokenAddress);
    }

    async getALBTTokenLegacy() {
        const albtAddress = await this.legacyRouterContract.albtToken();
        const albtDetails = await this.getTokenDetails(albtAddress);
        return { address: albtAddress, ...albtDetails };
    }

    async getALBTToken() {
        const albtAddress = await this.erc20RouterContract.albtToken();
        const albtDetails = await this.getTokenDetails(albtAddress);
        return { address: albtAddress, ...albtDetails };
    }

    async getProxyFeeLegacy(address) {
        return await this.legacyRouterProxyContract.feeAmountByToken(address)
    }

    async getFeeTokens() {
        return await this.erc20RouterContract.feeTokens();
    }

    async getFeeAmountByToken(tokenAddress) {
        return await this.erc20RouterContract.feeAmountByToken(tokenAddress);
    }

    fixSignatures(signatures) {
        const fixedSignatures = [];
        for (let i = 0; i < signatures.length; i++) {
            fixedSignatures[i] = signatures[i].startsWith("0x") ? signatures[i] : "0x" + signatures[i];
        }

        return fixedSignatures;
    }

    _getAllValidators(validators) {
        let validatorArray = [];
        for (const key in validators) {
            if (Object.hasOwnProperty.call(validators, key)) {
                const element = validators[key];
                validatorArray.push(element)
            }
        }
        return validatorArray;
    }

    async _setValidatorBasedOnVersion(validatorArray, checkedVersion) {
        const checkedMajorVersion = parseInt(checkedVersion.split('.')[0]);
        const checkedMinorVersion = parseInt(checkedVersion.split('.')[1]);
        for (let index = 0; index < validatorArray.length; index++) {
            let version
            let majorVersion
            let minorVersion
            try {
                version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                majorVersion = parseInt(version.split('.')[0]);
                minorVersion = parseInt(version.split('.')[1]);
            } catch (e) {
                continue;
            }
            if (majorVersion === checkedMajorVersion && minorVersion >= checkedMinorVersion) {
                this.validatorVersion = version;
                this.setValidator(validatorArray[index])
                break;
            }
        }
    }

    async _setNftValidatorBasedOnVersion(validatorArray, checkedVersion) {
        const checkedMajorVersion = parseInt(checkedVersion.split('.')[0]);
        const checkedMinorVersion = parseInt(checkedVersion.split('.')[1]);
        for (let index = 0; index < validatorArray.length; index++) {
            let version
            let majorVersion
            let minorVersion
            try {
                version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                majorVersion = parseInt(version.split('.')[0]);
                minorVersion = parseInt(version.split('.')[1]);
            } catch (e) {
                continue;
            }
            if (majorVersion === checkedMajorVersion && minorVersion >= checkedMinorVersion) {
                this.nftValidatorVersion = version;
                this.setNftValidator(validatorArray[index])
                break;
            }
        }
    }


    async _getDefaultConfiguredValidatorVersion() {
        const validatorArray = this._getAllValidators(this.config.validators);
        for (let index = 0; index < validatorArray.length; index++) {
            try {
                let version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                this.setValidator(validatorArray[index])
                return version;
            } catch (e) {
                continue;
            }
        }
    }

    async _getDefaultConfiguredNftValidatorVersion() {
        const validatorArray = this._getAllValidators(this.config.nftValidators);
        for (let index = 0; index < validatorArray.length; index++) {
            try {
                let version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                this.setNftValidator(validatorArray[index])
                return version;
            } catch (e) {
                continue;
            }
        }
    }

    async getUserCollections(userAddress, openseaApiKey) {
        try {

            const result = (await axios.get(`https://${network(this.chainId)}.opensea.io/api/v1/collections?asset_owner=${userAddress}&offset=0&limit=300`, {
                headers: {
                    "X-API-KEY": openseaApiKey
                }
            })).data;
            return result;
        } catch (e) {
            throw new Error(e.message)
        }
    }


    async getUserAssets(userAddress, assetContractAddress, openseaApiKey, nftTransferPage) {

        const offsetValue = 20;
        let offset = (nftTransferPage - 1) * offsetValue;
        try {
            const result = (await axios
                .get(`https://${network(this.chainId)}.opensea.io/api/v1/assets?owner=${userAddress}&asset_contract_address=${assetContractAddress}&order_direction=desc&offset=${offset}`, {
                    headers: {
                        "X-API-KEY": openseaApiKey
                    }
                })).data.assets;
            return result
        } catch (e) {
            throw new Error(e.message)
        }
    }

    async loadTokenMetadata(metadataUrl) {
        const domain = metadataUrl.split(":")[0];
        if (domain == 'ipfs' || domain == 'ipns') {
            try {
                const result = (await axios.get(uriToHttp(metadataUrl)[0])).data;
                return result;
            } catch (error) {
                throw new Error(e.message);
            }
        } else {
            try {
                const result = (await axios.get(metadataUrl)).data;
                return result
            } catch (e) {
                throw new Error(e.message)
            }
        }
    }

    async loadTokenByAddressAndId(collectionAddress, id, userAddress, skipOwnerCheck = false) {
        try {
            let tokenMetadata;
            const tokenContract = new ethers.Contract(collectionAddress, this.WrappedERC721Abi, this.signerOrProvider);

            if (!skipOwnerCheck) {
                let owner;
                try {
                    owner = await tokenContract.ownerOf(id);
                } catch (e) {
                    return;
                }
                if (owner.toLowerCase() !== userAddress.toLowerCase()) {
                    return;
                }
            }

            const collectionName = await tokenContract.name();
            const tokenURI = await tokenContract.tokenURI(id);
            try {
                tokenMetadata = await this.loadTokenMetadata(tokenURI);
            } catch (e) {
                console.error('Token Metadata cannot be loaded');
            }

            return { collectionName, ...tokenMetadata, id, tokenURI, collectionAddress }
        } catch (e) {
            throw new Error(e.message)
        }

    }

    async loadTokensByCollectionAddress(collectionAddress) {
        try {
            const tokenContract = new ethers.Contract(collectionAddress, this.WrappedERC721Abi, this.signerOrProvider);
            const userAddress = await this.signerOrProvider.getAddress();
            let balance;
            let collectionName;
            try {
                balance = await tokenContract.balanceOf(userAddress);
                collectionName = await tokenContract.name();
            } catch (e) {
                throw new Error('Cannot find collection at address.')
            }
            const tokens = {
                collectionName: '',
                collectionAddress: collectionAddress,
                items: []
            };
            tokens.collectionName = collectionName;
            if (balance.gt(0)) {
                for (let index = 0; index < balance.toString(); index++) {
                    let tokenMetadata;
                    let tokenID;
                    let tokenURI;
                    try {
                        tokenID = await tokenContract.tokenOfOwnerByIndex(userAddress, index);
                        tokenURI = await await tokenContract.tokenURI(tokenID);
                    } catch (e) {
                        throw new Error('Cannot find tokens by collection address only. Please, enter token ID.')
                    }

                    try {
                        const slicedDataLength = (tokenURI.slice(tokenURI.indexOf(",") + 1, tokenURI.length)).length;
                        const slicedDataLastSymbol = tokenURI.slice(tokenURI.indexOf(",") + 1, tokenURI.length)[slicedDataLength - 1];
                        const isEmbeddedData = (tokenURI.slice(tokenURI.indexOf(",") + 1, tokenURI.length))[0] === "{" && slicedDataLastSymbol === "}";
                        if (isEmbeddedData) {
                            tokenMetadata = JSON.parse((tokenURI.slice(tokenURI.indexOf(",") + 1, tokenURI.length)));
                        } else {
                            tokenMetadata = await this.loadTokenMetadata(tokenURI);
                        }
                    } catch (e) {
                        console.error('Token Metadata cannot be loaded');
                    }
                    tokens.items.push({ collectionName, ...tokenMetadata, id: tokenID.toString(), tokenURI })
                }
            }
            return tokens;
        } catch (e) {
            throw new Error(e.message)
        }

    }

    async approveNfts(collectionAddress) {
        const tokenContract = new ethers.Contract(collectionAddress, this.WrappedERC721Abi, this.signerOrProvider);
        return tokenContract.populateTransaction.setApprovalForAll(this.config.contracts.nftRouter, true);
    }

    async estimateApproveNfts(collectionAddress) {
        const tokenContract = new ethers.Contract(collectionAddress, this.WrappedERC721Abi, this.signerOrProvider);
        return tokenContract.estimateGas.setApprovalForAll(this.config.contracts.nftRouter, true);
    }

    async isNftsApproved(userAddress, collectionAddress) {
        const tokenContract = new ethers.Contract(collectionAddress, this.WrappedERC721Abi, this.signerOrProvider);
        return tokenContract.isApprovedForAll(userAddress, this.config.contracts.nftRouter);
    }

    async sendNftTokens(targetChainId, assets, userAddress, fee) {
        const nftRouter = new ethers.Contract(this.config.contracts.nftRouter, this.NFTRouterAbi, this.signerOrProvider);
        const assetInfo = [{ collection: assets[0].asset_contract.address, tokenIds: assets.map(x => BigNumber.from(x.token_id)) }]
        return nftRouter.populateTransaction.egress(targetChainId, assetInfo, userAddress, { value: fee });
    }

    async estimateSendNftTokens(targetChainId, assets, userAddress, fee) {
        const nftRouter = new ethers.Contract(this.config.contracts.nftRouter, this.NFTRouterAbi, this.signerOrProvider);
        const assetInfo = [{ collection: assets[0].asset_contract.address, tokenIds: assets.map(x => BigNumber.from(x.token_id)) }]
        return nftRouter.estimateGas.egress(targetChainId, assetInfo, userAddress, { value: fee });
    }

    async getNFTBridgeFee() {
        const nftRouter = new ethers.Contract(this.config.contracts.nftRouter, this.NFTRouterAbi, this.signerOrProvider);
        return nftRouter.serviceFee()
    }

    _decodeErc20Transmissions(transmissions) {
        const transferedTokens = [];
        for (let index = 0; index < transmissions.length; index++) {
            const transmissionDetails = transmissions[index];
            const payload = transmissionDetails.payload;
            let header, envelope;
            try {
                [header, envelope] = ethers.utils.defaultAbiCoder.decode(["uint256", "bytes"], payload)
            } catch (error) {
                console.error("The wrong payload is skipped", payload);
                continue;
            }
            let blob, subblob;
            let decoded = {};
            [decoded.sender, decoded.receiver, blob, decoded.deliveryFee] = ethers.utils.defaultAbiCoder.decode(["bytes", "bytes", "bytes", "uint256"], envelope);
            let action = {};
            [action.operation, action.token, subblob]
                = ethers.utils.defaultAbiCoder.decode(["uint8", "bytes", "bytes"], blob);

            if (action.operation == 0) { // burnUnlock
                let d = ethers.utils.defaultAbiCoder.decode(["uint256"], subblob);
                action.data = {
                    amount: d[0]
                };
            }
            else { // lockMint or burnMint
                let d = ethers.utils.defaultAbiCoder.decode(["uint256", "uint8", "string", "string", "uint8"], subblob);
                action.data = {
                    amount: d[0],
                    nativeChainId: d[1],
                    name: d[2],
                    symbol: d[3],
                    decimals: d[4],
                };
            }
            decoded.action = action;
            transferedTokens.push({ decoded, transmissionDetails })
        }
        return transferedTokens;
    }

    _decodeNftTransmissions(transmissions) {
        const collections = [];
        for (let index = 0; index < transmissions.length; index++) {
            const transmissionDetails = transmissions[index];
            const payload = transmissionDetails.payload;
            let header, envelope;
            try {
                [header, envelope] = ethers.utils.defaultAbiCoder.decode(["uint256", "bytes"], payload)
            } catch (error) {
                console.error("The wrong payload is skipped", payload);
                continue;
            }
            let blob, subblob;
            let decoded = {};
            [decoded.sender, decoded.receiver, blob] = ethers.utils.defaultAbiCoder.decode(["bytes", "bytes", "bytes[]"], envelope);
            decoded.actions = [];
            for (let i in blob) {
                let action = {};
                [action.operation, action.collection, subblob]
                    = ethers.utils.defaultAbiCoder.decode(["uint8", "bytes", "bytes"], blob[i]);
                if (action.operation == 0) { // unlock
                    let d = ethers.utils.defaultAbiCoder.decode(["uint256[]"], subblob);
                    action.data = {
                        tokenIds: d[0],
                    };
                }
                else { // mint
                    let d = ethers.utils.defaultAbiCoder.decode(["uint256[]", "uint8", "string[]", "string", "string"], subblob);
                    action.data = {
                        tokenIds: d[0],
                        nativeChainId: d[1],
                        tokenURIs: d[2],
                        name: d[3],
                        symbol: d[4],
                    };
                }
                decoded.actions[i] = action;
            }
            collections.push({ decoded, transmissionDetails })
        }
        return collections;
    }

    async getErc20Claims(targetChainId, userAddress) {
        const response = await this.backendAxios(`/erc20/claims?targetChainId=${targetChainId}&receiver=${userAddress}`, { responseType: "json" });
        const transmissions = response.data.map((item) => {
            item.delivered = item.delivery ? true : false;
            return item;
        });;

        return this._decodeErc20Transmissions(transmissions);
    }

    async getNftClaims(targetChainId, userAddress) {
        const response = await this.backendAxios(`/erc721/claims?targetChainId=${targetChainId}&receiver=${userAddress}`, { responseType: "json" });
        const transmissions = response.data.map((item) => {
            item.delivered = item.delivery ? true : false;
            return item;
        });;

        const decodedTransmissions = this._decodeNftTransmissions(transmissions);
        const transmissionsForMulticall = [];
        const nftData = [];

        for (let index = 0; index < decodedTransmissions.length; index++) {
            const decodedTransmission = decodedTransmissions[index];
            const operation = decodedTransmission.decoded.actions[0].operation;
            const collectionAddress = decodedTransmission.decoded.actions[0].collection;
            const tokenIds = decodedTransmission.decoded.actions[0].data.tokenIds;
            if (operation != 0) {
                const collectionName = decodedTransmission.decoded.actions[0].data.name;
                const tokenURIs = decodedTransmission.decoded.actions[0].data.tokenURIs;
                for (let i = 0; i < tokenIds.length; i++) {
                    const tokenURI = tokenURIs[i];
                    let tokenMetadata;
                    try {
                        tokenMetadata = await this.loadTokenMetadata(tokenURI);
                    } catch (e) {
                        console.error('Token Metadata cannot be loaded');
                    }
                    nftData.push({ collectionAddress, collectionName, tokenURI, ...tokenMetadata, ...decodedTransmission.transmissionDetails, receiver: decodedTransmission.decoded.receiver })
                }
            } else {
                transmissionsForMulticall.push({ collectionAddress: collectionAddress, tokenIds: tokenIds, transmissionDetails: decodedTransmission.transmissionDetails, receiver: decodedTransmission.decoded.receiver })
            }
        }
        const multicallNfts = await this._loadNftsByCollectionAndIdMulticall(transmissionsForMulticall)
        const concatedArray = nftData.concat(multicallNfts);

        return concatedArray;
    }

    async estimateErc20AutoClaim(
        sourceChainId,      // number
        targetChainId,      // number
        tokenAddress,       // address, like 0x...
        amount,             // string, like 0.1
        senderAddress,      // address, like 0x...
        receiverAddress,    // address, like 0x...
    ) {
        const response = await this.backendAxios(
            `/erc20/estimateDeliveryFee?sourceChainId=${sourceChainId}&targetChainId=${targetChainId}&token=${tokenAddress}&amount=${amount}&sender=${senderAddress}&receiver=${receiverAddress}`,
            { responseType: "json" });

        return response.data; // { fee: <ethers.BigNumber>, expiry: <number>, signature: <hex string> }
    }

    async _loadNftsByCollectionAndIdMulticall(collections) {
        const tokenInterface = new ethers.utils.Interface(this.WrappedERC721Abi);
        const fragment_name = tokenInterface.getFunction('name');
        const fragment_tokenURI = tokenInterface.getFunction('tokenURI');

        const multi = new MultiCall(this.signerOrProvider);
        const inputs = [];
        const nftData = [];
        let tokenData = [];
        if (this.chainId == 1287 || this.chainId == 1284 || this.chainId == 421611 || this.chainId == 42161 || this.chainId == 1313161555) {
            for (let collection of collections) {
                inputs.push({ target: collection.collectionAddress, callData: tokenInterface.encodeFunctionData(fragment_name, []) })
                for (let id of collection.tokenIds) {
                    inputs.push({ target: collection.collectionAddress, callData: tokenInterface.encodeFunctionData(fragment_tokenURI, [id]) })
                }
            }

            tokenData = (await this.multicallContract.callStatic.tryBlockAndAggregate(false, inputs)).returnData;

        } else {
            for (let collection of collections) {
                inputs.push({ target: collection.collectionAddress, function: 'name', args: [] })
                for (let id of collection.tokenIds) {
                    inputs.push({ target: collection.collectionAddress, function: 'tokenURI', args: [id] })
                }
            }
            const chunkSize = 20;
            for (let i = 0; i < inputs.length; i += chunkSize) {
                const chunk = inputs.slice(i, i + chunkSize);
                tokenData.push((await multi.multiCall(this.WrappedERC721Abi, chunk))[1]);
            }
            tokenData = [].concat.apply([], tokenData)
        }


        let indexName = 0;
        for (let i = 0; i < collections.length; i++) {
            const collectionName = tokenData[indexName];
            for (let j = indexName + 1; j < collections[i].tokenIds.length + indexName + 1; j++) {
                const tokenURI = tokenData[j];
                let tokenMetadata;
                try {
                    tokenMetadata = await this.loadTokenMetadata(tokenURI);
                } catch (e) {
                    console.error('Token Metadata cannot be loaded');
                }
                const data = {
                    collectionName: collectionName,
                    tokenURI: tokenURI,
                    ...tokenMetadata,
                    ...collections[i].transmissionDetails,
                    receiver: collections[i].receiver
                }
                nftData.push(data);
            }
            indexName += collections[i].tokenIds.length + 1;
        }
        return nftData;
    }

    async claim(transmission) {
        const teleportContract = new ethers.Contract(this.config.contracts.teleport, this.ITeleportAbi, this.signerOrProvider);
        return teleportContract.populateTransaction.deliver(
            transmission.transmissionId,
            transmission.sourceChainId,
            transmission.transmissionSender,
            transmission.transmissionReceiver,
            transmission.dAppId,
            transmission.payload,
            transmission.signatures);
    }

    async estimateClaim(transmission) {
        const teleportContract = new ethers.Contract(this.config.contracts.teleport, this.ITeleportAbi, this.signerOrProvider);
        return teleportContract.estimateGas.deliver(
            transmission.transmissionId,
            transmission.sourceChainId,
            transmission.transmissionSender,
            transmission.transmissionReceiver,
            transmission.dAppId,
            transmission.payload,
            transmission.signatures)
    }
}

export { RouterSDK };